import * as vscode from "vscode";
import * as child_process from "child_process";

const GitFilter: { [key: string]: string } = {
  s: "since",
  u: "until",
  a: "author",
  g: "grep",
};
export interface GitCommit {
  hash?: string;
  author?: string;
  date?: string;
  msg?: string;
  files?: string[];
}

export interface GitCommitFilter{
  path?:string;
  contains?:string;
  committer?:string;
  dateRange?:[Date, Date]
}



export class GitService {
  private _repoPath: string;

  constructor(private gitPath: string, private logger: vscode.LogOutputChannel) {
    const workspaceFolders = vscode.workspace.workspaceFolders;
    if (!workspaceFolders || workspaceFolders.length === 0) {
      throw new Error("No workspace folder opened");
    }
    this._repoPath = workspaceFolders[0].uri.fsPath;
  }

  private transformFilter(filter: GitCommitFilter): string[] {
    const gitArgs: string[] = [];
    if(filter.committer) {
      gitArgs.push(`--author=${filter.committer}`);
    }
    if(filter.dateRange) {
      const [from, to] = filter.dateRange;
      gitArgs.push(`--since=${this.getDateStr(from)}`);
      // get next day of to
      const toNextDay = new Date(to);
      toNextDay.setDate(toNextDay.getDate() + 1);
      gitArgs.push(`--until=${this.getDateStr(toNextDay)}`);
    }
    if(filter.contains) {
      gitArgs.push(`--grep=${filter.contains}`);
    }
    return gitArgs;
  }

  getDateStr(d:Date) :string {
    return d.toLocaleDateString('zh-CN');
  }
  private defaultFilter(): string[] {
    return ['--author=$(git config user.name)', '--since=\"7 days ago\"'];
  }
  getCommitters() : string[] {
    const command = `git log --format='%aN' | sort -u`;
    const result = child_process.execSync(`git log --format='%aN' | sort -u`, {
      cwd: this._repoPath,
    });
    return result.toString().split("\n");
  }

  getCurrentUser(): string {
    const result = child_process.execSync(`git config user.name`, {
      cwd: this._repoPath,
    });
    return result.toString().trim();
  }

  public async getCommits(filter: string = ""): Promise<any[]> {
    const filterParts = filter.split(" ");
    let nFilter: string[] = [];
    filterParts.forEach((part) => {
      const t = part.split(":");
      const key = t[0];
      const v = t[1];
      if (key in GitFilter) {
        nFilter.push(`--${GitFilter[key]}=${v}`);
      }
    });
    if(nFilter.length === 0){
      nFilter = this.defaultFilter();
    }
    const gitArgs = [
      "log",
      "--oneline",
      ...nFilter,
      '--pretty=format:"{\\"hash\\":\\"%h\\",\\"author\\":\\"%an\\",\\"date\\":\\"%ad\\",\\"msg\\":\\"%s\\"}"%n---',
      "--date=format:%Y-%m-%d",
      "--name-only",
    ];
 
    const that = this;
    return new Promise((resolve, reject) => {
      let lines: string[] = [];
      let commits: GitCommit[] = [];

      function makeCommit(lines: any[]): GitCommit | undefined {
        try {
          if(lines.length === 1) {
            return undefined;
          }
            const commitLine = lines[0];
            that.logger.info('add commit',commitLine);
           
          const commit: GitCommit = JSON.parse(lines[0]);
          commit.files = [];
          for (let i = 1; i < lines.length; i++) {
            if (lines[i].trim() !== "") {
              commit.files?.push(lines[i]);
            }
          }
          return commit;
        } catch (e) {
          console.log(e);
        }

        return undefined;
      }
      const gitProcess = child_process.spawn(this.gitPath, gitArgs, {
        cwd: this._repoPath,
        shell: true,
      });
      gitProcess.stderr.setEncoding("utf8");
      gitProcess.stderr.on("data", (data) => {
        console.log(`stderr: ${data}`);
      });
      gitProcess.stdout.setEncoding("utf8");
      gitProcess.on("error", (error) => {
        console.log(`error: ${error.message}`);
        reject(error);
      });
      gitProcess.stdout.on("data", (data) => {
        const nlines = data.split("\n").filter((l:string) => l.trim() !== "");
        for (let nline of nlines) {
          that.logger.info('process line 1',nline);
          if (nline.startsWith("{")) {
            if(lines.length > 0){
              const commit = makeCommit(lines);
              if(commit){
                that.logger.info('push commit', commit.hash)
                commits.push(commit);
               
              }
            }
            lines = [];
            lines.push(nline);
          } else if (nline.startsWith("---") ) {
            continue
          }else if( nline.trim() === ""){
            const commit = makeCommit(lines);
            if(commit){
              that.logger.info('push commit 2', commit.hash)
              commits.push(commit);
            }
            lines = [];
          } else if (nline.trim() !== "") {
            lines.push(nline);
          }
        }
       
      });
      gitProcess.on("exit",(code,signal)=>{
        if(code === 0){
          if(lines.length > 0){
            const commit = makeCommit(lines);
            if(commit){

              that.logger.info('push commit 3', commit.hash)
              commits.push(commit);
            }
          }
          resolve(commits);
        }else{
            reject(new Error(`git log exited with code ${code}`));
        }
      });
    });
  }

  public async getPreviousCommit(hash: string): Promise<string> {
    const command = `git log --pretty=format:"%h" -n 1 ${hash}^`;
    return new Promise((resolve, reject) => {
      child_process.exec(
        command,
        { cwd: this._repoPath },
        (error, stdout, stderr) => {
          if (error) {
            reject(error);
          } else {
            resolve(stdout.trim());
          }
        }
      );
    });
  }
}
